package experiments;

import core.Solution;
import core.app.App;
import core.processingchain.actions.policies.pool.PushPolicy;
import core.processingchain.actions.pool.NotificationAction;
import core.processingchain.actions.pool.PushAction;
import core.processingchain.metadata.MetadataType;
import edu.ucc.core.Simulation;
import edu.ucc.core.events.simulationevents.ContentReceivedEvent;
import edu.ucc.network.architecture.ArchitectureParameters;
import edu.ucc.statisticscollection.StatisticsCalculator;
import edu.ucc.utils.FileConstants;
import edu.ucc.utils.Logging;
import loader.AbstractSystemLoader;
import loader.SystemLoader;
import sampleapps.platerecognition.*;
import workload.mobility.MobilityParameters;
import workload.requests.ContentRequestParameters;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static edu.ucc.utils.FileConstants.KILOBYTE;
import static edu.ucc.utils.FileConstants.MEGABYTE;
import static edu.ucc.utils.TimeConstants.MINUTE;
import static edu.ucc.utils.TransmissionConstants.KILOMETER;
import static utils.ScenarioConstants.*;
import static utils.Utils.createGenericMobilityParameters;

public class IoTProcessingLauncher {
    final int appId = 1;
    final String appName = "app1";
    final String baseURI = "www.app1.com/";

    private App createAppUpLink() {
        App appUpLink = new App(appId, appName, baseURI, FileConstants.GIGABYTE);
        appUpLink.addMetadataExtractor(MetadataType.CONTENT_METADATA, new ContentMetadataExtractorPlateRecognition());

        appUpLink.addPreProcessor(ContentReceivedEvent.class, new PreProcessorPlateRecognition());
        appUpLink.addEventDetector(PPOutcomePlateRecognition.class, new EventDetectorPlateRecognition());
        appUpLink.addMetadataExtractor(MetadataType.INTERNAL_EVENT_METADATA, new EventMetadataExtractorPlateRecognition());

        appUpLink.addActionForEventType(PlateRecognitionEvent.class, new NotificationAction());
        appUpLink.setExternalSubscriberPerEventType(PlateRecognitionEvent.class,
                List.of("http://www.policedepartment.com", "http://www.fineticketingsystem.com"));
        appUpLink.addActionForEventType(PlateRecognitionEvent.class, new PushAction());
        appUpLink.setPolicyForActionType(PushAction.class, new PushPolicy(false, true, false));
        return appUpLink;
    }

    private ContentRequestParameters createContentParametersForAppUplink(double requestEndTime) {
        // This is for pull workload.requests
        final int numFileRequests = 0;
        final int numOfFiles = 0;
        final long minFileSize = MEGABYTE;
        final long maxFileSize = MEGABYTE; // 2 * MEGABYTE;

        final int numPullDataRequests = 0;
        final long minDataSizeInPullDataRequests = 0;
        final long maxDataSizeInPullDataRequests = 0;

        // This is for push workload.requests
        final int numPushDataRequests = 100;
        final long minDataSizeInPushDataRequests = 500 * KILOBYTE;
        final long maxDataSizeInPushDataRequests = 5 * MEGABYTE;
        return new ContentRequestParameters(appId, baseURI, numFileRequests, numOfFiles, minFileSize,
                maxFileSize, numPullDataRequests, minDataSizeInPullDataRequests,
                maxDataSizeInPullDataRequests, numPushDataRequests, minDataSizeInPushDataRequests, maxDataSizeInPushDataRequests,
                0.0, requestEndTime);
    }

    private ArchitectureParameters createNetworkArchitectureParameters() {
        return new ArchitectureParameters(
                BS_RANGE,
                KILOMETER * 100, KILOMETER * 100,
                TOTAL_UPLINK_BW_FOR_CLOUD_CHILDREN, TOTAL_DOWNLINK_BW_FOR_CLOUD_CHILDREN,
                TOTAL_UPLINK_BW_FOR_ES_CHILDREN, TOTAL_DOWNLINK_BW_FOR_ES_CHILDREN, TOTAL_BW_FOR_ES_SIBLINGS,
                TOTAL_UPLINK_BW_FOR_BS_CHILDREN, TOTAL_DOWNLINK_BW_FOR_BS_CHILDREN,
                NUM_EDGE_SERVERS, STORAGE_PER_EDGE_SERVER,
                600, MTU);
    }

    private void executeImmediateProcessingScenario(String outputDirPath) {
        // 1. Define edu.ucc.network's architecture
        ArchitectureParameters architectureParameters = createNetworkArchitectureParameters();
        MobilityParameters mobilityParameters = createGenericMobilityParameters();
        mobilityParameters.setMobilityEnabled(false);

        // 2. Define workload per individual app
        final ContentRequestParameters contentRequestParameters = createContentParametersForAppUplink(mobilityParameters.getTrajectoryTime());

        Map<Integer, ContentRequestParameters> parametersPerApp = new HashMap<>();
        parametersPerApp.put(1, contentRequestParameters);

        // 3. Create apps, specifying/selecting the preprocessors, event detectors and policies.
        App iotApp = createAppUpLink();

        // 4. Create the solution, linking it with the testbed
        AbstractSystemLoader systemLoader = new SystemLoader(architectureParameters, mobilityParameters, parametersPerApp);
        systemLoader.loadSystem();
        final Solution solution = new Solution(systemLoader.getNetworkArchitecture());

        // 5. Register apps with the solution
        solution.registerApp(iotApp);

        Logging.configureLogging(Logging.WARNING | Logging.CONTENT_RECEIVED | Logging.INFO | Logging.PUSH_DATA_REQUEST);
        final Simulation simulation = Simulation.getInstance();

        // 6. Load testbed's queue with the events of workload ()workload.requests + workload.mobility
        simulation.init(systemLoader.getNetworkArchitecture(), systemLoader.getWorkload(), solution, TRAJECTORY_TIME + MINUTE);

        // 7. Launch simulation
        simulation.startSimulation();

        // 8. Analyse results by the statistics collector / calculator
        StatisticsCalculator statisticsCalculator = new StatisticsCalculator(simulation.getStatisticsCollector(), simulation.getWorkload(), simulation.getRunningTime());
        final int numRequests = systemLoader.getWorkload().getRequestEvents().size();
        final int numOfFiles = contentRequestParameters.getNumOfFiles();
        final int numOfUploads = contentRequestParameters.getNumPushDataRequests();
        statisticsCalculator.writeStatistics(outputDirPath, architectureParameters.getNumUserEquipments(), numRequests, numOfFiles, numOfUploads);
        simulation.restart();
    }

    private void executeScheduledProcessingScenario(String outputDirPath) {
        // 1. Define edu.ucc.network's architecture
        ArchitectureParameters architectureParameters = createNetworkArchitectureParameters();
        MobilityParameters mobilityParameters = createGenericMobilityParameters();
        mobilityParameters.setMobilityEnabled(false);

        // 2. Define workload per individual app
        final ContentRequestParameters contentRequestParameters = createContentParametersForAppUplink(mobilityParameters.getTrajectoryTime());

        Map<Integer, ContentRequestParameters> parametersPerApp = new HashMap<>();
        parametersPerApp.put(1, contentRequestParameters);

        // 3. Create apps, specifying/selecting the preprocessors, event detectors and policies.
        App iotApp = createAppUpLink();

        // Set the processing schedule
        iotApp.setProcessingInterval(3600.0);

        // 4. Create the solution, linking it with the testbed
        AbstractSystemLoader systemLoader = new SystemLoader(architectureParameters, mobilityParameters, parametersPerApp);
        systemLoader.loadSystem();
        final Solution solution = new Solution(systemLoader.getNetworkArchitecture());

        // 5. Register apps with the solution
        solution.registerApp(iotApp);

        Logging.configureLogging(Logging.WARNING | Logging.CONTENT_RECEIVED | Logging.INFO | Logging.PUSH_DATA_REQUEST);
        final Simulation simulation = Simulation.getInstance();

        // 6. Load testbed's queue with the events of workload ()workload.requests + workload.mobility
        simulation.init(systemLoader.getNetworkArchitecture(), systemLoader.getWorkload(), solution, TRAJECTORY_TIME + MINUTE);

        // 7. Launch simulation
        simulation.startSimulation();

        // 8. Analyse results by the statistics collector / calculator
        StatisticsCalculator statisticsCalculator = new StatisticsCalculator(simulation.getStatisticsCollector(), simulation.getWorkload(), simulation.getRunningTime());
        final int numRequests = systemLoader.getWorkload().getRequestEvents().size();
        final int numOfFiles = contentRequestParameters.getNumOfFiles();
        final int numOfUploads = contentRequestParameters.getNumPushDataRequests();
        statisticsCalculator.writeStatistics(outputDirPath, architectureParameters.getNumUserEquipments(), numRequests, numOfFiles, numOfUploads);
        simulation.restart();
    }

    public static void main(String[] args) {
        String outputDirPath = System.getProperty("user.home") + File.separator + "Desktop" + File.separator + "iot-data-processing-immediate" + File.separator;
        new IoTProcessingLauncher().executeImmediateProcessingScenario(outputDirPath);

        outputDirPath = System.getProperty("user.home") + File.separator + "Desktop" + File.separator + "iot-data-processing-scheduled" + File.separator;
        new IoTProcessingLauncher().executeScheduledProcessingScenario(outputDirPath);
    }
}
